#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include "lb-VdW.h"
#include "lb-messages.h"

double
lb_VdW_pressure_2d(const LBVdWParameters* params,
		     const double* LB_RESTRICT rho,
		     const double* LB_RESTRICT drx,
		     const double* LB_RESTRICT dry,
		     const double* LB_RESTRICT ddr)
{
	lb_assert(params && rho && drx && dry && ddr);

	return (*rho)*(params->T/(1.0 - params->b*(*rho))
		     - params->a*(*rho))
		- params->K*((*drx)*(*drx) + (*dry)*(*dry) + 0.5*(*rho)*(*ddr));
}

void
lb_d2q9_VdW_equilibrium(const LBVdWParameters* params,
			  const double* LB_RESTRICT averages,
			  const double* LB_RESTRICT drx,
			  const double* LB_RESTRICT dry,
			  const double* LB_RESTRICT ddr,
			  double* LB_RESTRICT eq)
{
	const double rho = *averages++;

	const double ux = *averages++;
	const double uy = *averages;

	const double rux = rho*ux;
	const double ruy = rho*uy;
	const double rux2 = rux*ux;
	const double ruy2 = ruy*uy;

	const double drx2 = params->K*(*drx)*(*drx);
	const double dry2 = params->K*(*dry)*(*dry);
	const double rddr = params->K*rho*(*ddr);

	const double brho = params->b*rho;
	const double brho1 = 1.0 - brho;
	const double Td = params->T/brho1;
	const double arho = params->a*rho;

	const double p0 = rho*(Td - arho);
	const double trP = (p0 - rddr)/3;
	const double dp0 = Td*(1.0 + brho/brho1) - 2*arho;
	const double nu = (params->tau - 0.5)*(1.0/3.0 - dp0);
	const double nuudr = nu*(ux*(*drx) + uy*(*dry));

	{
		const double t1 = 0.25*(drx2 - dry2);
		const double t2 = - trP + 0.5*nuudr;
		const double t3 = - nu*uy*(*dry) + t1 - t2 + (rux2 - 0.5*ruy2)/3;
		const double t4 = - nu*ux*(*drx) - t1 - t2 + (ruy2 - 0.5*rux2)/3;
		const double t5 = rux/3.0;
		const double t6 = ruy/3.0;

		*eq++ = t3 + t5;
		*eq++ = t4 + t6;
		*eq++ = t3 - t5;
		*eq++ = t4 - t6;
	}

	{
		const double t1 = 0.25*nu*(ux*(*dry) + uy*(*drx));
		const double t2 = nuudr + 0.25*trP + (rux2 + ruy2)/12;
		const double t3 = t2 + t1;
		const double t4 = t2 - t1;
		const double t5 = (rux + ruy)/12;
		const double t6 = (ruy - rux)/12;
		const double t7 = 0.25*(rux*uy + params->K*(*drx)*(*dry));

		*eq++ = t3 + t5 + t7;
		*eq++ = t4 + t6 - t7;
		*eq++ = t3 - t5 + t7;
		*eq++ = t4 - t6 - t7;
	}

	*eq = rho - 5*trP - 2*(rux2 + ruy2)/3;
}

double
lb_VdW_pressure_3d(const LBVdWParameters* params,
		     const double* LB_RESTRICT rho,
		     const double* LB_RESTRICT drx,
		     const double* LB_RESTRICT dry,
		     const double* LB_RESTRICT drz,
		     const double* LB_RESTRICT ddr)
{
	lb_assert(params && rho && drx && dry && drz && ddr);

	return (*rho)*(params->T/(1.0 - params->b*(*rho))
		     - params->a*(*rho))
		- params->K*((*drx)*(*drx) + (*dry)*(*dry) + (*drz)*(*drz)
			    + 0.5*(*rho)*(*ddr));
}

void
lb_d3q19_VdW_equilibrium(const LBVdWParameters* params,
			   const double* LB_RESTRICT rhop,
			   const double* LB_RESTRICT up,
			   const double* LB_RESTRICT drx,
			   const double* LB_RESTRICT dry,
			   const double* LB_RESTRICT drz,
			   const double* LB_RESTRICT ddr,
			   double* LB_RESTRICT eq)
{
	const double rho = *rhop;

	const double ux = *up++;
	const double uy = *up++;
	const double uz = *up;

	const double rux = rho*ux;
	const double ruy = rho*uy;
	const double ruz = rho*uz;

	const double rux2 = rux*ux;
	const double ruy2 = ruy*uy;
	const double ruz2 = ruz*uz;

	const double drx2 = params->K*(*drx)*(*drx);
	const double dry2 = params->K*(*dry)*(*dry);
	const double drz2 = params->K*(*drz)*(*drz);

	const double brho = params->b*rho;
	const double brho1 = 1.0 - brho;
	const double Td = params->T/brho1;
	const double arho = params->a*rho;
	const double p0 = rho*(Td - arho);

	const double Pcc = p0 - params->K*rho*(*ddr);
	const double Pxx = Pcc + 0.5*(drx2 - dry2 - drz2);
	const double Pxy = params->K*(*drx)*(*dry);
	const double Pxz = params->K*(*drx)*(*drz);
	const double Pyy = Pcc + 0.5*(dry2 - drx2 - drz2);
	const double Pyz = params->K*(*dry)*(*drz);
	const double Pzz = Pcc + 0.5*(drz2 - drx2 - dry2);

	const double trP = (Pxx + Pyy + Pzz)/9.0;
	const double ru2 = rux2 + ruy2 + ruz2;

	const double dp0 = Td*(1.0 + brho/brho1) - 2*arho;
	const double nu = (params->tau - 0.5)*(1.0/3.0 - dp0);

	const double ndrx = nu*(*drx);
	const double ndry = nu*(*dry);
	const double ndrz = nu*(*drz);

	const double nuudr = ux*ndrx + uy*ndry + uz*ndrz;

	{
		const double t1 = 0.25*(ru2 - trP) - nuudr;

		{
			const double t2 = t1 + 0.25*(Pxx + rux2) + 0.5*ndrx*ux;
			const double t3 = rux/6.0;

			eq[0] = t2 + t3;
			eq[1] = t2 - t3;
		}

		{
			const double t2 = t1 + 0.25*(Pyy + ruy2) + 0.5*ndry*uy;
			const double t3 = ruy/6.0;

			eq[2] = t2 + t3;
			eq[3] = t2 - t3;
		}

		{
			const double t2 = t1 + 0.25*(Pzz + ruz2) + 0.5*ndrz*uz;
			const double t3 = ruz/6.0;

			eq[4] = t2 + t3;
			eq[5] = t2 - t3;
		}
	}

	{
		const double t1 = 0.5*(0.5*nuudr - trP - 0.25*ru2);

		{
			const double t2 = t1 + 0.125*(Pxx + Pyy);

			{
				const double t3 = 0.25*(ux + uy);
				const double t4 = rux + ruy;
				const double t5 = t2 + 0.5*t3*t4 
						+ t3*(ndrx + ndry) + 0.25*Pxy;
				const double t6 = t4/12.0;

				eq[6] = t5 + t6;
				eq[9] = t5 - t6;
			}

			{
				const double t3 = 0.25*(ux - uy);
				const double t4 = rux - ruy;
				const double t5 = t2 + 0.5*t3*t4
						+ t3*(ndrx - ndry) - 0.25*Pxy;
				const double t6 = t4/12.0;

				eq[7] = t5 - t6;
				eq[8] = t5 + t6;
			}
		}

		{
			const double t2 = t1 + 0.125*(Pxx + Pzz);

			{
				const double t3 = 0.25*(ux + uz);
				const double t4 = rux + ruz;
				const double t5 = t2 + 0.5*t3*t4 
						+ t3*(ndrx + ndrz) + 0.25*Pxz;
				const double t6 = t4/12.0;

				eq[10] = t5 + t6;
				eq[13] = t5 - t6;
			}

			{
				const double t3 = 0.25*(ux - uz);
				const double t4 = rux - ruz;
				const double t5 = t2 + 0.5*t3*t4
						+ t3*(ndrx - ndrz) - 0.25*Pxz;
				const double t6 = t4/12.0;

				eq[11] = t5 - t6;
				eq[12] = t5 + t6;
			}
		}

		{
			const double t2 = t1 + 0.125*(Pyy + Pzz);

			{
				const double t3 = 0.25*(uy + uz);
				const double t4 = ruy + ruz;
				const double t5 = t2 + 0.5*t3*t4 
						+ t3*(ndry + ndrz) + 0.25*Pyz;
				const double t6 = t4/12.0;

				eq[14] = t5 + t6;
				eq[17] = t5 - t6;
			}

			{
				const double t3 = 0.25*(uy - uz);
				const double t4 = ruy - ruz;
				const double t5 = t2 + 0.5*t3*t4
						+ t3*(ndry - ndrz) - 0.25*Pyz;
				const double t6 = t4/12.0;

				eq[15] = t5 - t6;
				eq[16] = t5 + t6;
			}
		}
	}

	eq[18] = rho - 6.0*trP - 1.5*ru2;
}
