#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <lua.h>
#include <lauxlib.h>

#include "lb-messages.h"
#include "lblua-main.h"

#ifdef LB_ENABLE_MPI
#include <mpi.h>
#include "lb-mpi-utils.h"
#endif /* LB_ENABLE_MPI */

#ifndef LBLUA_PROMPT1
#define LBLUA_PROMPT1 ">"
#endif /* LBLUA_PROMPT1 */

#ifndef LBLUA_PROMPT2
#define LBLUA_PROMPT2 ">>"
#endif /* LBLUA_PROMPT2 */

static int
lblua_readline(lua_State* ls, const char* prompt)
{
	static char buffer[512];

	lb_assert(ls != NULL);

	fputs(prompt, stdout);
	fflush(stdout);

	if (fgets(buffer, sizeof(buffer) - 1, stdin) == NULL)
		return 0;  /* read fails */
	else {
		lua_pushstring(ls, buffer);
		return 1;
	}
}

static int
lblua_line_complete(lua_State* ls, int status)
{
	lb_assert(ls != NULL);

	if (status == LUA_ERRSYNTAX &&
		strstr(lua_tostring(ls, -1), "near `<eof>'") != NULL) {

		lua_pop(ls, 1);
		return 0;
	} else
		return 1;
}

static int
lblua_loadstring(lua_State* ls)
{
	int status;

	lb_assert(ls != NULL);

	/* clear stack */
	lua_settop(ls, 0);

	if (lblua_readline(ls, LBLUA_PROMPT1) == 0)
		return -1;

	for (;;) {  /* repeat until gets a complete line */
		status = luaL_loadbuffer(ls, lua_tostring(ls, 1),
					 lua_strlen(ls, 1), "=stdin");

		if (lblua_line_complete(ls, status))
			break;

		if (lblua_readline(ls, LBLUA_PROMPT2) == 0)
    			return -1;

		lua_concat(ls, lua_gettop(ls));  /* join lines */
	}

	lua_remove(ls, 1);  /* remove line */
	return status;
}

static void
lblua_interactive(lua_State* ls)
{
	int status;

	lb_assert(ls != NULL);

	while ((status = lblua_loadstring(ls)) != -1) {
		if (status == 0) {
			/* get traceback function */
			lua_pushliteral(ls, "_TRACEBACK");
			lua_rawget(ls, LUA_GLOBALSINDEX);

			lb_assert(lua_isfunction(ls, -1));

			/* put it under the chunk */
			lua_insert(ls, 1);

			status = lua_pcall(ls, 0, 0, 1);

			/* remove traceback function */
			lua_remove(ls, 1);
		}

		if (status != 0) {
			const char* msg = lua_tostring(ls, -1);

			if (msg == NULL)
				msg = "(error with no message)";

			fprintf(stderr, "*** %s\n", msg);
			lua_pop(ls, 1);
		}
	}

	/* clear stack */
	lua_settop(ls, 0);
	fputs("\n", stdout);
}

static void
lblua_file(lua_State* ls, const char* filename)
{
	int status;

	lb_assert(ls != NULL);

	if (filename) {

		/*
		 *  check if file exist && is readable
		 */

		{
			struct stat buf;

			if (stat(filename, &buf) < 0)
				lb_fatal("cannot stat `%s': %s\n",
					    filename, strerror(errno));

			if (!S_ISREG(buf.st_mode))
				lb_fatal("`%s' is not a regular file\n",
					    filename);

			if (access(filename, R_OK) < 0)
				lb_fatal("cannot read `%s': %s\n",
					    filename, strerror(errno));
		}
	}

	/*
	 *  execute file
	 */

	status = luaL_loadfile(ls, filename);

	if (status == 0) {
		/* get traceback function */
		lua_pushliteral(ls, "_TRACEBACK");
		lua_rawget(ls, LUA_GLOBALSINDEX);

		lb_assert(lua_isfunction(ls, -1));

		/* put it under the chunk */
		lua_insert(ls, 1);

		status = lua_pcall(ls, 0, LUA_MULTRET, 1);

		/* remove traceback function */
		lua_remove(ls, 1);
	}

	if (status) {
		const char* msg = lua_tostring(ls, -1);

		if (msg == NULL)
			msg = "(error with no message)";
#ifdef LB_ENABLE_MPI
		if(lb_mpi_comm_rank(MPI_COMM_WORLD) == 0)
#endif /* LB_ENABLE_MPI */
			fprintf(stderr, "*** %s\n", msg);

		lua_pop(ls, 1);
	}
}

static void
lblua_prepare_args(lua_State* ls, int argc, char** argv)
{
	lb_assert(ls != NULL);
	lb_assert(argc > 0 && argv != NULL && argv[0] != NULL);

	if (argc > 1) {
		int a;

		lua_newtable(ls);

		for (a = 1; a < argc; ++a) {
			lua_pushnumber(ls, a - 1);
			lua_pushstring(ls, argv[a]);
			lua_rawset(ls, -3);
		}

		/*
		 *  arg.n = maximum index in table `arg'
		 */

		lua_pushliteral(ls, "n");
		lua_pushnumber(ls, argc - 2);
		lua_rawset(ls, -3);

		lua_setglobal(ls, "arg");
	}
}

static void
lblua_openlibs(lua_State* ls, const lua_CFunction* libs)
{
	lb_assert(ls != NULL);

	if (libs != NULL) {
		for (; *libs != NULL; ++libs) {
			(*libs)(ls);
			/* discard any results */
			lua_settop(ls, 0);
		}
	}
}

void
lblua_main(int argc, char** argv, const lua_CFunction* libs)
{
	lua_State* ls;

	lb_assert(argc > 0);
	lb_assert(argv != NULL);

	ls = lua_open();
	if (ls == NULL) {
		lb_fatal("cannot create lua_State\n");
		return;
	}

	lblua_openlibs(ls, libs);

	if (argc == 1 && isatty(0)) {
#ifdef LB_ENABLE_MPI
		if (lb_mpi_comm_size(MPI_COMM_WORLD) > 1) {
			if (lb_mpi_comm_rank(MPI_COMM_WORLD) == 0)
				fprintf(stderr, "*** interactive session"
				        " possible only with one MPI "
					"process ***\n");
			goto out;
		}
#endif /* LB_ENABLE_MPI */
		lblua_interactive(ls);
	} else {
		lblua_prepare_args(ls, argc, argv);
		lblua_file(ls, (argc > 1 ? argv[1] : NULL));
	}

#ifdef LB_ENABLE_MPI
out:
#endif /* LB_ENABLE_MPI */

	lua_close(ls);
}
