#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <lua.h>
#include <lauxlib.h>

#include "lb-macros.h"
#include "lb-memory.h"
#include "lb-messages.h"

#include "lb-rgb.h"
#include "lua-lb-rgb.h"

#include "lblua-class.h"
#include "lblua-utils.h"

#ifndef LUA_LB_RGB_CLASSNAME
#define LUA_LB_RGB_CLASSNAME "rgb"
#endif /* LUA_LB_RGB_CLASSNAME */

#ifndef LUA_LB_COLORMAP_CLASSNAME
#define LUA_LB_COLORMAP_CLASSNAME "colormap"
#endif /* LUA_LB_RGB_COLORMAP_CLASSNAME */

static void
push_color(lua_State* ls, const double color[3])
{
	lb_assert(ls != NULL);
	lb_assert(color != NULL);

	lua_pushnumber(ls, (lua_Number) color[0]);
	lua_pushnumber(ls, (lua_Number) color[1]);
	lua_pushnumber(ls, (lua_Number) color[2]);
}

static void
get_color(lua_State* ls, int i, double color[3])
{
	double tmp;

	lb_assert(ls != NULL);
	lb_assert(color != NULL);

	tmp = (double) luaL_checknumber(ls, i++);
	color[0] = LB_CLAMP(tmp, 0.0, 1.0);

	tmp = (double) luaL_checknumber(ls, i++);
	color[1] = LB_CLAMP(tmp, 0.0, 1.0);

	tmp = (double) luaL_checknumber(ls, i++);
	color[2] = LB_CLAMP(tmp, 0.0, 1.0);
}

/*
 *  LBColormap first
 */

static int
lua_lb_colormap_new(lua_State* ls)
{
	LBColormap* map;

	lb_assert(ls != NULL);
	lblua_checkargcount(ls, 0, LUA_LB_COLORMAP_CLASSNAME"()");

	map = lb_colormap_new();

	lua_boxpointer(ls, map);
	luaL_getmetatable(ls, LUA_LB_COLORMAP_CLASSNAME);
	lua_setmetatable(ls, -2);

	return 1;
}

static int
lua_lb_colormap_destroy(lua_State* ls)
{
	LBColormap* map;

	lb_assert(ls != NULL);
	map = lblua_checkinstance(LBColormap, ls, 1,
				    LUA_LB_COLORMAP_CLASSNAME);
	lb_colormap_destroy(map);

	return 0;
}

static int
lua_lb_colormap_typename(lua_State* ls)
{
	lb_assert(ls != NULL);

	lblua_checkargcount(ls, 1, LUA_LB_COLORMAP_CLASSNAME":typename()");
	(void) lblua_checkinstance(LBColormap, ls, 1,
				     LUA_LB_COLORMAP_CLASSNAME);
	lua_pushstring(ls, LUA_LB_COLORMAP_CLASSNAME);

	return 1;
}

static int
lua_lb_colormap_num_colors(lua_State* ls)
{
	LBColormap* map;

	lb_assert(ls != NULL);
	lblua_checkargcount(ls, 1, LUA_LB_COLORMAP_CLASSNAME":num_colors()");
	map = lblua_checkinstance(LBColormap, ls, 1,
				    LUA_LB_COLORMAP_CLASSNAME);
	lua_pushnumber(ls, (lua_Number) lb_colormap_num_colors(map));

	return 1;
}

static int
lua_lb_colormap_get_color(lua_State* ls)
{
	int c;
	LBColormap* map;

	lb_assert(ls != NULL);
	lblua_checkargcount(ls, 2, LUA_LB_COLORMAP_CLASSNAME
				     ":get_color(int)");
	map = lblua_checkinstance(LBColormap, ls, 1,
				    LUA_LB_COLORMAP_CLASSNAME);
	c = luaL_checkint(ls, 2);

	if (c >= 0 && c < lb_colormap_num_colors(map)) {
		double color[3];

		lb_colormap_get_color(map, c, color);
		push_color(ls, color);

		return 3;
	} else
		return lblua_range_error(ls, LUA_LB_COLORMAP_CLASSNAME
					       ":get_color(int)");
}

static int
lua_lb_colormap_set_color(lua_State* ls)
{
	int c;
	LBColormap* map;

	lb_assert(ls != NULL);
	lblua_checkargcount(ls, 5, LUA_LB_COLORMAP_CLASSNAME
				     ":set_color(int,double,double,double)");
	map = lblua_checkinstance(LBColormap, ls, 1,
				    LUA_LB_COLORMAP_CLASSNAME);
	c = luaL_checkint(ls, 2);

	if (c >= 0 && c < lb_colormap_num_colors(map)) {
		double color[3];

		get_color(ls, 3, color);
		lb_colormap_set_color(map, c, color);

		return 0;
	} else
		return lblua_range_error(ls, LUA_LB_COLORMAP_CLASSNAME
				       ":set_color(int,double,double,double)");
}

static int
lua_lb_colormap_append_color(lua_State* ls)
{
	LBColormap* map;

	lb_assert(ls != NULL);
	lblua_checkargcount(ls, 4, LUA_LB_COLORMAP_CLASSNAME
				     ":append_color(double,double,double)");
	map = lblua_checkinstance(LBColormap, ls, 1,
				    LUA_LB_COLORMAP_CLASSNAME);

	{
		double color[3];

		get_color(ls, 2, color);
		lb_colormap_append_color(map, color);
	}

	return 0;
}

static int
lua_lb_colormap_map_value(lua_State* ls)
{
	double v;
	LBColormap* map;

	lb_assert(ls != NULL);
	lblua_checkargcount(ls, 2, LUA_LB_COLORMAP_CLASSNAME
				     ":map_value(double)");
	map = lblua_checkinstance(LBColormap, ls, 1,
				    LUA_LB_COLORMAP_CLASSNAME);

	if (lb_colormap_num_colors(map) < 2)
		luaL_error(ls, LUA_LB_COLORMAP_CLASSNAME
			       ":map_value(double): colormap should contain"
			       " at least two colors");

	v = (double) luaL_checknumber(ls, 2);

	{
		double color[3];

		lb_colormap_map_value(map, v, color);
		push_color(ls, color);
	}

	return 3;
}

static void
lua_lb_colormap_register(lua_State* ls, const char* modulename)
{
	const luaL_reg colormap_methods[] = {
		{"typename",     lua_lb_colormap_typename},
		{"num_colors",   lua_lb_colormap_num_colors},
		{"get_color",    lua_lb_colormap_get_color},
		{"set_color",    lua_lb_colormap_set_color},
		{"append_color", lua_lb_colormap_append_color},
		{"map_value",    lua_lb_colormap_map_value},
		{NULL, NULL}
	};

	lb_assert(ls != NULL);

	lblua_register_class(ls, modulename,
			       LUA_LB_COLORMAP_CLASSNAME,
			       lua_lb_colormap_new,
			       lua_lb_colormap_destroy,
			       colormap_methods);
}

/*
 *  LBRGB
 */

static int
lua_lb_rgb_new(lua_State* ls)
{
	int w, h;
	LBRGB* rgb;

	lb_assert(ls != NULL);

	lblua_checkargcount(ls, 2, LUA_LB_RGB_CLASSNAME"(int,int)");

	w = luaL_checkint(ls, 1);
	h = luaL_checkint(ls, 2);

	if (w <= 0 || h <= 0)
		luaL_error(ls, "%s(int,int) expects positive numbers",
			       LUA_LB_RGB_CLASSNAME);

	rgb = lb_rgb_new(w, h);

	lua_boxpointer(ls, rgb);
	luaL_getmetatable(ls, LUA_LB_RGB_CLASSNAME);
	lua_setmetatable(ls, -2);

	return 1;
}

static int
lua_lb_rgb_destroy(lua_State* ls)
{
	LBRGB* rgb;

	lb_assert(ls != NULL);
	rgb = lblua_checkinstance(LBRGB, ls, 1, LUA_LB_RGB_CLASSNAME);
	lb_rgb_destroy(rgb);

	return 0;
}

static int
lua_lb_rgb_typename(lua_State* ls)
{
	lb_assert(ls != NULL);

	lblua_checkargcount(ls, 1, LUA_LB_RGB_CLASSNAME":typename()");
	(void) lblua_checkinstance(LBRGB, ls, 1, LUA_LB_RGB_CLASSNAME);
	lua_pushstring(ls, LUA_LB_RGB_CLASSNAME);

	return 1;
}

static int
lua_lb_rgb_width(lua_State* ls)
{
	LBRGB* rgb;

	lb_assert(ls != NULL);
	lblua_checkargcount(ls, 1, LUA_LB_RGB_CLASSNAME":width()");
	rgb = lblua_checkinstance(LBRGB, ls, 1, LUA_LB_RGB_CLASSNAME);

	lua_pushnumber(ls, lb_rgb_width(rgb));
	return 1;
}

static int
lua_lb_rgb_height(lua_State* ls)
{
	LBRGB* rgb;

	lb_assert(ls != NULL);
	lblua_checkargcount(ls, 1, LUA_LB_RGB_CLASSNAME":height()");
	rgb = lblua_checkinstance(LBRGB, ls, 1, LUA_LB_RGB_CLASSNAME);

	lua_pushnumber(ls, lb_rgb_height(rgb));
	return 1;
}

static int
lua_lb_rgb_fill(lua_State* ls)
{
	LBRGB* rgb;
	double color[3];

	lb_assert(ls != NULL);
	lblua_checkargcount(ls, 4, LUA_LB_RGB_CLASSNAME
				     ":fill(double,double,double)");
	rgb = lblua_checkinstance(LBRGB, ls, 1, LUA_LB_RGB_CLASSNAME);

	get_color(ls, 2, color);
	lb_rgb_fill(rgb, color);

	return 0;
}

static int
lua_lb_rgb_get_pixel(lua_State* ls)
{
	LBRGB* rgb;
	int x, y;

	lb_assert(ls != NULL);
	lblua_checkargcount(ls, 3, LUA_LB_RGB_CLASSNAME":get_pixel(int,int)");
	rgb = lblua_checkinstance(LBRGB, ls, 1, LUA_LB_RGB_CLASSNAME);

	x = luaL_checkint(ls, 2);
	y = luaL_checkint(ls, 3);

	if (x >= 0 && x < lb_rgb_width(rgb) &&
	    y >= 0 && y < lb_rgb_height(rgb)) {
		double pixel[3];

		lb_rgb_get_pixel(rgb, x, y, pixel);
		push_color(ls, pixel);

		return 3;
	} else
		return lblua_range_error(ls, LUA_LB_RGB_CLASSNAME
					       ":get_pixel(int,int)");
}

static int
lua_lb_rgb_set_pixel(lua_State* ls)
{
	LBRGB* rgb;
	int x, y;

	lb_assert(ls != NULL);

	lblua_checkargcount(ls, 6, LUA_LB_RGB_CLASSNAME
				     ":set_pixel(int,int,double,double,double)");
	rgb = lblua_checkinstance(LBRGB, ls, 1, LUA_LB_RGB_CLASSNAME);

	x = luaL_checkint(ls, 2);
	y = luaL_checkint(ls, 3);

	if (x >= 0 && x < lb_rgb_width(rgb) &&
	    y >= 0 && y < lb_rgb_height(rgb)) {
		double pixel[3];

		get_color(ls, 4, pixel);
		lb_rgb_set_pixel(rgb, x, y, pixel);

		return 0;
	} else
		return lblua_range_error(ls, LUA_LB_RGB_CLASSNAME
					   ":set_pixel(int,int,double,double,double)");
}

static int
lua_lb_rgb_set_pixel_rgba(lua_State* ls)
{
	LBRGB* rgb;
	int x, y;

	lb_assert(ls != NULL);

	lblua_checkargcount(ls, 7, LUA_LB_RGB_CLASSNAME
				     ":set_pixel_rgba(int,int,"
				     "double,double,double,double)");
	rgb = lblua_checkinstance(LBRGB, ls, 1, LUA_LB_RGB_CLASSNAME);

	x = luaL_checkint(ls, 2);
	y = luaL_checkint(ls, 3);

	if (x >= 0 && x < lb_rgb_width(rgb) &&
	    y >= 0 && y < lb_rgb_height(rgb)) {
		lua_Number tmp;
		double pixel[4];

		get_color(ls, 4, pixel);

		tmp = luaL_checknumber(ls, 7);
		pixel[3] = LB_CLAMP(tmp, 0.0, 255.0);

		lb_rgb_set_pixel_rgba(rgb, x, y, pixel);

		return 0;
	} else
		return lblua_range_error(ls, LUA_LB_RGB_CLASSNAME
				":set_pixel_rgba(int,int,"
				"double,double,double,double)");
}

static int
lua_lb_rgb_map_value(lua_State* ls)
{
	LBRGB* rgb;
	LBColormap* cmap;
	int x, y;

	lb_assert(ls != NULL);

	lblua_checkargcount(ls, 5, LUA_LB_RGB_CLASSNAME
				     ":map_value(colormap,int,int,double)");
	rgb = lblua_checkinstance(LBRGB, ls, 1, LUA_LB_RGB_CLASSNAME);
	cmap = lblua_checkinstance(LBColormap, ls, 2,
				     LUA_LB_COLORMAP_CLASSNAME);

	if (lb_colormap_num_colors(cmap) < 2)
		luaL_error(ls, LUA_LB_RGB_CLASSNAME
			       ":map_value(colormap,int,int,double): "
			       "colormap should contain at least two colors");

	x = luaL_checkint(ls, 3);
	y = luaL_checkint(ls, 4);

	if (x >= 0 && x < lb_rgb_width(rgb) &&
	    y >= 0 && y < lb_rgb_height(rgb)) {
		lb_rgb_map_value(rgb, cmap, x, y, luaL_checknumber(ls, 5));

		return 0;
	} else
		return lblua_range_error(ls, LUA_LB_RGB_CLASSNAME
					   ":map_value(colormap,int,int,double)");
}

static int
lua_lb_rgb_save(lua_State* ls)
{
	const char* filename;
	LBRGB* rgb;

	lb_assert(ls != NULL);
	lblua_checkargcount(ls, 2, LUA_LB_RGB_CLASSNAME":save()");
	rgb = lblua_checkinstance(LBRGB, ls, 1, LUA_LB_RGB_CLASSNAME);

	filename = luaL_checkstring(ls, 2);

	{
		FILE* file;

		file = fopen(filename, "wb");
		if (file == NULL)
			return luaL_error(ls, "failed to open `%s' for writing",
					      filename);

		lb_rgb_save(rgb, file);
		fclose(file);
	}

	return 0;
}

void
lua_lb_rgb_register(lua_State* ls, const char* modulename)
{
	const luaL_reg rgb_methods[] = {
		{"typename",       lua_lb_rgb_typename},
		{"width",          lua_lb_rgb_width},
		{"height",	   lua_lb_rgb_height},
		{"fill",           lua_lb_rgb_fill},
		{"get_pixel",      lua_lb_rgb_get_pixel},
		{"set_pixel",      lua_lb_rgb_set_pixel},
		{"set_pixel_rgba", lua_lb_rgb_set_pixel_rgba},
		{"map_value",      lua_lb_rgb_map_value},
		{"save",           lua_lb_rgb_save},
		{NULL, NULL}
	};

	lb_assert(ls != NULL);

	lua_lb_colormap_register(ls, modulename);

	lblua_register_class(ls, modulename, LUA_LB_RGB_CLASSNAME,
			       lua_lb_rgb_new, lua_lb_rgb_destroy,
			       rgb_methods);
}
